//
// file: primitive3D_vertex_nrm.fx
// version: v1.5
// author: Ren712
//

//--------------------------------------------------------------------------------------
// Settings
//--------------------------------------------------------------------------------------
float3 sElementPosition = float3(0, 0, 0);
float3 sElementRotation = float3(0, 0, 0);
float4 sElementColor = float4(0, 0, 0, 0);

float2 sElementSize = float2(1, 1);
bool sFlipTexture = false;
bool sFlipVertex = false;
bool sIsBillboard = false;
bool sFogEnable = false;
int fCullMode = 1;

float2 uvMul = float2(1, 1);
float2 uvPos = float2(0, 0);

float sDispHeight = 0.5;

float sSpecularPower = 4;
float3 sSpecularColor = float3(1, 1, 1);
float3 sLightDir = float3(0.507, -0.507, -0.2);
float3 nStrength = float3(0.5, 0.5, 0.5);

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------
texture sTexColor;
texture sTexDisplace;

//--------------------------------------------------------------------------------------
// Variables set by MTA
//--------------------------------------------------------------------------------------
float4x4 gWorldViewProjection : WORLDVIEWPROJECTION;
float4x4 gProjection : PROJECTION;
float4x4 gView : VIEW;
float4x4 gViewInverse : VIEWINVERSE;
float3 gCameraPosition : CAMERAPOSITION;
float gTime : TIME;
int gCapsMaxAnisotropy < string deviceCaps="MaxAnisotropy"; >;
int CUSTOMFLAGS < string skipUnusedParameters = "yes"; >;

//--------------------------------------------------------------------------------------
// Sampler 
//--------------------------------------------------------------------------------------
sampler2D SamplerColor = sampler_state
{
    Texture = (sTexColor);
    MipFilter = Linear;
    MaxAnisotropy = gCapsMaxAnisotropy;
    MinFilter = Anisotropic;
    AddressU = Mirror;
    AddressV = Mirror;
};

sampler SamplerDisplace = sampler_state
{
    texture = (sTexDisplace);
    MinFilter = Point;
    MagFilter = Point;
    MipFilter = None;
    AddressU = Clamp;
    AddressV = Clamp;
};

//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------
struct VSInput
{
    float3 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float4 Diffuse : COLOR0;
};

struct PSInput
{
    float4 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float4 WorldPos : TEXCOORD1;
    float3 WorldNormal : TEXCOORD2;
    float2 Depth : TEXCOORD3;
    float4 Diffuse : COLOR0;
};

//--------------------------------------------------------------------------------------
// Create world matrix with world position and euler rotation
//--------------------------------------------------------------------------------------
float4x4 createWorldMatrixEuler(float3 pos, float3 rot)
{
    float4x4 eleMatrix = {
        float4(cos(rot.z) * cos(rot.y) - sin(rot.z) * sin(rot.x) * sin(rot.y), 
                cos(rot.y) * sin(rot.z) + cos(rot.z) * sin(rot.x) * sin(rot.y), -cos(rot.x) * sin(rot.y), 0),
        float4(-cos(rot.x) * sin(rot.z), cos(rot.z) * cos(rot.x), sin(rot.x), 0),
        float4(cos(rot.z) * sin(rot.y) + cos(rot.y) * sin(rot.z) * sin(rot.x), sin(rot.z) * sin(rot.y) - 
                cos(rot.z) * cos(rot.y) * sin(rot.x), cos(rot.x) * cos(rot.y), 0),
        float4(pos.x,pos.y,pos.z, 1),
    };
    return eleMatrix;
}

//--------------------------------------------------------------------------------------
// Create world matrix with world position and vector
//--------------------------------------------------------------------------------------
float4x4 createWorldMatrixVector( float3 pos, float3 fwVec, float3 upVec )
{
    float3 zaxis = normalize( fwVec );    // The "forward" vector.
    float3 xaxis = normalize( cross( -upVec, zaxis ));// The "right" vector.
    float3 yaxis = cross( xaxis, zaxis );     // The "up" vector.

    // Create a 4x4 world matrix from the right, up, forward and eye position vectors
    float4x4 worldMatrix = {
        float4( xaxis.x,  xaxis.y, xaxis.z, 0 ),
        float4( yaxis.x,  yaxis.y, yaxis.z, 0 ),
        float4( zaxis.x,  zaxis.y, zaxis.z, 0 ),
        float4(	  pos.x,    pos.y,   pos.z, 1 )
    };
    return worldMatrix;
}

//--------------------------------------------------------------------------------------
// Vertex Shader 
//--------------------------------------------------------------------------------------
PSInput VertexShaderFunctionSM3(VSInput VS)
{
    PSInput PS = (PSInput)0;

    // set proper position and scale of the quad
    VS.Position.xyz = float3(- 0.5 + VS.TexCoord.xy, 0);
	
    if (!sFlipTexture) VS.TexCoord.y = 1 - VS.TexCoord.y;

    // get displacement texture
    float2 texCoord = (VS.TexCoord * uvMul) + uvPos;
    float4 dispTex = tex2Dlod(SamplerDisplace, float4(texCoord, 0.0, 0.0));

    // get normal vector
    dispTex.xyz = normalize((dispTex.xyz * 2.0) - 1.0);

    // add displacement
    VS.Position.z -= sDispHeight * dispTex.a;
	
    // set scale of the quad
    VS.Position.xy *= sElementSize.xy;
	
    // flip vertex position xyz - xzy
    if ((sFlipVertex) && (!sIsBillboard)) VS.Position.xyz = VS.Position.xzy;
	
    // create WorldMatrix for the quad
    float4x4 sWorld;
    if (sIsBillboard) sWorld = createWorldMatrixVector(sElementPosition, gViewInverse[2].xyz, gViewInverse[1].xyz);
        else sWorld = createWorldMatrixEuler(sElementPosition, sElementRotation);
	
    // calculate screen position of the vertex
    float4 wPos = mul(float4( VS.Position, 1), sWorld);
    float4 vPos = mul(wPos, gView);
    PS.Depth = float2(vPos.z, vPos.w);
    PS.Position = mul(vPos, gProjection);

    // pass texCoords and vertex color to PS
    PS.TexCoord = (VS.TexCoord * uvMul) + uvPos;
    PS.Diffuse = VS.Diffuse * sElementColor;

    // set normal vector
    float3 Normal = normalize(dispTex.xyz * nStrength);
	
    // set information for surface normal and lighting calculation in PS	
    PS.WorldPos = mul(float4(VS.Position,1), sWorld);
    PS.WorldNormal = normalize(mul(Normal, (float3x3)sWorld));

    return PS;
}

PSInput VertexShaderFunctionSM2(VSInput VS)
{
    PSInput PS = (PSInput)0;

    // set proper position and scale of the quad	
    VS.Position.xyz = float3(- 0.5 + VS.TexCoord.xy, 0);
    VS.Position.xy *= sElementSize.xy;
    if (!sFlipTexture) VS.TexCoord.y = 1 - VS.TexCoord.y;
    if ((sFlipVertex) && (!sIsBillboard)) VS.Position.xyz = VS.Position.xzy;
	
    // create WorldMatrix for the quad
    float4x4 sWorld;
    if (sIsBillboard) sWorld = createWorldMatrixVector(sElementPosition, gViewInverse[2].xyz, gViewInverse[1].xyz);
        else sWorld = createWorldMatrixEuler(sElementPosition, sElementRotation);
	
    // calculate screen position of the vertex
    float4 wPos = mul(float4( VS.Position, 1), sWorld);
    float4 vPos = mul(wPos, gView);
    PS.Position = mul(vPos, gProjection);

    // pass texCoords and vertex color to PS
    PS.TexCoord = (VS.TexCoord * uvMul) + uvPos;
    PS.Diffuse = VS.Diffuse * sElementColor;

    // set normal vector
    float3 Normal = float3(0,0,1);
	
    // set information for surface normal and lighting calculation in PS	
    PS.WorldPos = mul(float4(VS.Position,1), sWorld);
    PS.WorldNormal = normalize(mul(Normal, (float3x3)sWorld));
	
    return PS;
}

//------------------------------------------------------------------------------------------
// MTAApplyFog
//------------------------------------------------------------------------------------------
int gFogEnable                     < string renderState="FOGENABLE"; >;
float4 gFogColor                   < string renderState="FOGCOLOR"; >;
float gFogStart                    < string renderState="FOGSTART"; >;
float gFogEnd                      < string renderState="FOGEND"; >;

float3 MTAApplyFog( float3 texel, float distFromCam )
{
    if ( !gFogEnable )
        return texel;
    float FogAmount = ( distFromCam - gFogStart )/( gFogEnd - gFogStart );
    texel.rgb = lerp(texel.rgb, gFogColor.rgb, saturate( FogAmount) );
    return texel;
}

//--------------------------------------------------------------------------------------
// Pixel shaders 
//--------------------------------------------------------------------------------------
float4 PixelShaderFunctionSM3(PSInput PS) : COLOR0
{
    // sample color texture
    float4 finalColor = tex2D(SamplerColor, PS.TexCoord.xy);

    // calculate specular light
    float3 lightDir = normalize(sLightDir);

    float3 h = normalize(normalize(gCameraPosition - PS.WorldPos) - lightDir);
    float specLighting = pow(saturate(dot(h, PS.WorldNormal)), sSpecularPower);

    // Stop underneath artifacts
    float lightAwayDot = -dot(lightDir, PS.WorldNormal);
    specLighting *= saturate(3 *  lightAwayDot);
    specLighting = saturate(specLighting);	
		
    // add specular lighting to color texture
    finalColor.rgb += finalColor.rgb * specLighting * sSpecularColor.rgb;
	
    // multiply by vertex color
    finalColor *= PS.Diffuse;
	
    // apply fog effect
    finalColor.rgb = MTAApplyFog(finalColor.rgb, PS.Depth.x / PS.Depth.y);

    return saturate(finalColor);
}

float4 PixelShaderFunctionSM2(PSInput PS) : COLOR0
{
    // sample color texture
    float4 finalColor = tex2D(SamplerColor, PS.TexCoord.xy);

    // calculate specular light
    float3 lightDir = normalize(sLightDir);

    float3 h = normalize(normalize(gCameraPosition - PS.WorldPos) - lightDir);
    float specLighting = pow(saturate(dot(h, PS.WorldNormal)), sSpecularPower);

    // Stop underneath artifacts
    float lightAwayDot = -dot(lightDir, PS.WorldNormal);
    specLighting *= saturate(3 *  lightAwayDot);
    specLighting = saturate(specLighting);	
		
    // add specular lighting to color texture
    finalColor.rgb += finalColor.rgb * specLighting * sSpecularColor.rgb;
	
    // multiply by vertex color
    finalColor *= PS.Diffuse;

    return saturate(finalColor);
}

//--------------------------------------------------------------------------------------
// Techniques
//--------------------------------------------------------------------------------------
technique dxDrawPrimitive3Dfx_vertex_nrm_VS3
{
  pass P0
  {
    ZEnable = true;
    ZFunc = LessEqual;
    ZWriteEnable = true;
    CullMode = fCullMode;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = InvSrcAlpha;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_3_0 VertexShaderFunctionSM3();
    PixelShader  = compile ps_3_0 PixelShaderFunctionSM3();
  }
} 

technique dxDrawPrimitive3Dfx_vertex_nrm_VS2
{
  pass P0
  {
    ZEnable = true;
    ZFunc = LessEqual;
    ZWriteEnable = true;
    CullMode = fCullMode;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = InvSrcAlpha;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunctionSM2();
    PixelShader  = compile ps_2_0 PixelShaderFunctionSM2();
  }
}

// Fallback
technique fallback
{
  pass P0
  {
    // Just draw normally
  }
}
